#ifndef LIGHT_H
#define LIGHT_H

#include "WPIlib.h"
#include "IOmap.h"
#include "Storage.h"
#include "Random.h"

class Light {
public:
	typedef enum {ballPos, rider1, rider2, kinect1, kinect2, kinect3, blink1, binary, rainbow} LightShows;
	/* List of lightshows:
	 * ballPos: Show ball sensors on the lights.  Used for positioning balls during disabled
	 * rider1: Knightrider pattern, one color
	 * rider2: Knightrider pattern, changes color
	 * kinect1: Select light with kinect
	 * kinect2: Control color and select light with kinect 
	 */
	Light(Storage* storage, GenericHID* left, GenericHID* right);
	~Light();
	
	
	/*
	 * Control startup signalling.  NOTE: no other Light functions should
	 * be invoked between a showStartupWarning(true) and showStartupWarning(false).
	 * Startup warning is a warning light that gyro is being calibrated and
	 * the robot should not be disturbed. It occurs in hardware and does
	 * not need any process to be called (those SHOULD NOT be called as
	 * mentioned above)
	 */
	void showStartupWarning(bool enable);
	
	/* Name: processLights
	 * Arguments: None
	 * Actions: Checks the robot's current state 
	 * 			and updates the lights to match
	 * Returns: Nothing
	 */
	void processLights();
	
	/* Name: processLightShow
	 * Arguments: None
	 * Actions: Reads input from the Kinect and changes the lights based on input
	 * 		RUNNING THIS AT THE SAME TIME AS PROCESS LIGHTS WILL NOT WORK
	 * Returns: Nothing
	 */
	void processLightShow();
	
	/* Name: selectLightShow
	 * Arguments: pattern: The type of lightshow you want, stored in the enum lightShows
	 * Actions: Selects the pattern you want to run when calling processLightShow
	 * Returns: Nothing
	 */
	void selectLightShow(LightShows pattern);
	
	/* Name: nextPattern
	 * Arguments: None
	 * Actions: Selects the next pattern every time it's called
	 * Returns: Nothing
	 */
	void nextPattern();
	
	/* Name: setColor
	 * Arguments: float red: 0 to 1: set red component of lights
	 * 			  float green: 0 to 1: set green component of lights
	 * 			  float blue: 0 to 1: set blue component of lights
	 * Actions: Sets the color of the lights
	 * Returns: Nothing
	 */
	void setColor(float red, float green, float blue);
	
	/* Name: setLight
	 * Arguments: Which lights are on?
	 * Actions: Turns the lights on or off
	 * Returns: Nothing
	 */
	void setLight(bool light1, bool light2, bool light3);
	
	/* Name: setCoopertition
	 * Arguments: true: signal white lights, if possible
	 * 			  false: Don't signal
	 * Actions: tells the lights we want to display white lights if possible
	 * Returns: The clothes you bought yesterday
	 */
	void setCoopertition(bool signal);

	/* Name: setShooting
	 * Arguments: true: signal blue lights, if possible
	 * 			  false: Don't signal
	 * Actions: tells the lights we want to display blue lights if possible
	 * Returns: Nothing
	 */
	void setShooting(bool shootingStatus);
	
	/* Name: showBallPos
	 * Arguments: None
	 * Actions: Overrides the current lightshow and shows the status of the ball sensors duing disabled
	 * 		Useful for positioning balls before a match
	 * Returns: Nothing
	 */
	void showBallPos();
	
	/* Name: setAlliance
	 * Arguments: True for red alliance
	 * 			False for blue alliance
	 * Actions: Sets the color of the lights for some patterns
	 * Returns: Nothing
	 */
	void setAlliance(bool color);
	
private:
	DigitalOutput lightRed;
	DigitalOutput lightGreen;
	DigitalOutput lightBlue;
	DigitalOutput lightSelect1;
	DigitalOutput lightSelect2;
	DigitalOutput lightSelect3;
	
	GenericHID* kinectLeft;
	GenericHID* kinectRight;
	
	Timer timer;
	
	Storage* thunderStorage;
		
	bool shooting;
	bool signalRequest;
	bool requestNextPattern;
	bool alliance;
	
	int binaryCount;
	
	LightShows lightShow;
};

#endif //LIGHT_H
