#ifndef SHOOTER
#define SHOOTER

#include "Storage.h"
#include "ThunderCANJaguar.h"
#include "Encoder.h"
#include "PIDController.h"
#include "Timer.h"
#include "Counter.h"
#include "UDPLogger.h"
#include "MonoEncoder.h"

// THESE MIN/MAX HERE FOR CONTROLS 
//turret min/max in driver degrees (use to clamp desired input angle and 
// translate to/from turret position ticks)
#define TURRET_MAX_DEG 120.0
#define TURRET_MIN_DEG -200.0


//debug
#define SHOOTER_DEBUG 0
#define SHOOTER_LOGGING 1
#define TURRET_DEBUG 0

#define TOP_LOGGER_PORT 1511
#define BOTTOM_LOGGER_PORT 1512

class LimitedOutput : public PIDOutput {
public:
	LimitedOutput(PIDOutput *out, AnalogChannel *fwdLimit, AnalogChannel *revLimit);
	virtual void PIDWrite(float f);
private:
	PIDOutput *output;
	AnalogChannel *fwdLimit;
	AnalogChannel *revLimit;
};



class Shooter {
public:
	Shooter(Storage * s);
	virtual ~Shooter(); // virtual to shut up the warning

// Shoot -------------------------------------------------------
	
	/* Name: Shoot
	 * Arguments: none
	 * Actions: requests that the ball be shot (this will happen
	 *          once a] a ball is available; b] wheel motors are up
	 *          to speed; and c] the turret angle is stable)
	 * Returns: noithoingk
	 */
	virtual void shoot();

	/* Name: shootFinished
	 * Arguments: none
	 * Actions: checks if a ball has been shot
	 * Returns: true: ball shot
	 * 			false: ball not shot
	 */
	virtual bool shootFinished();
	
	/* Name: shootStop
	 * Arguments: none
	 * Actions: stops the shooting need 
	 * Returns: nothing
	 */
	virtual void shootStop();
	
// Process ------------------------------------------------------------
	
	/* Name: processShooter
	 * Arguments: none
	 * Actions: Keeps moving the turret toward the desired angle.
	 * 			Stops shooting when a ball has left. 
	 * Returns: nada
	 * 
	 * Intended to be called often.
	 */
	virtual void processShooter();
	// ^ use PID loop for turret

// Wheels -----------------------------------------------------------------

	/* Name: setSpeed
	 * Arguments: top speed: angular velocity for top shooter wheel
	 * 			bottom speed: angular velocity of bottom wheel
	 *        Speeds are in RPS if wheels are not marked broken
	 *        Speeds are motor powers (0 -> 1.0) if wheels are broken
	 * Actions: It sets the angular velocity of the shooter
	 * Returns: nothin'
	 */
	void setSpeed(float top_speed, float bottom_speed);// rots por soconds
	
	/* Name: setNormalizedSpeed
	 * Arguments: top speed: Speed for top shooter wheel
	 * 			  bottom speed: Speed for bottom shooter wheel
	 * 			Speeds are motor power (0..1)
	 * Actions: Sets the speed of the shooter wheels
	 * Returns: Nothing
	 */
	void setNormalizedSpeed(float top_speed, float bottom_speed);
	
	/* Name: atSpeed
	 * Arguments: none
	 * Actions: checks if the angular velocities for the wheels have been reached
	 * Returns: true: both wheels reach angular velocities
	 * 			false: one or both wheels too fast/slow 
	 */
	bool atSpeed();
	
	/* Name: setWheelsBroken
	 * Arguments: true: broken
	 * 			  false: working
	 * Actions: Marks the shooter wheel encoders as broken
	 * 			this makes the wheels move continously, with power as input
	 * Returns: nothin'
	 */
	void setWheelsBroken(bool broken);

	/* Name: setWheelsVariable
	 * Arguments: true: variable wheel speeds
	 * 			  false: continuous wheel speeds
	 * Actions: makes the shooter, iff the encoders work,
	 * 			either make the wheels spin constantly or
	 * 			make the wheels spin only when firing is wanted;
	 * 			variable saves battery, continuous allows almost instantaneous shooting
	 * Returns: nothing
	 * 
	 * 
	 */
	void setWheelsVariable(bool variable);
	
	/*Name: setUseBackspin
	 * Args: Enable backspin?
	 * Actions: Enables slowing down the top shooter to put backspin on the ball
	 * Returns: Nothing
	 */
	void setUseBackspin(bool enable);
	
	/* Name: getXSpeed
	 * Args: none
	 * Actions: finds the wheel angular velocities
	 * Returns: rotations per second of wheel in question
	 */
	float getTopSpeed(); // top roller
	float getBottomSpeed();// bottom roller
	
	/* Get raw encoder values */
	int getTopEnc();
	int getBottomEnc();

	/* get revs per sec */
	double getTopRev();
	double getBottomRev();
	
// Turret -----------------------------------------------------------

	// Question: how do we know the turret initial angle? giveTurretAngle()
	
	/* Name: setTurretAngle
	* Arguments: angle: turret angle in degrees
	* 					(from robot forward main axis, counterclockwise positive)
	* Actions: Sets the angle to which to rotate the turret
	* Returns: nothin'
	*/
	void setTurretAngle(float angle);
	
	/* Actions: sets the angle that the turret is currently at to a number
	 * Or just make the default turret angle be 0, and have things constrained
	 *  to the clockwisemost bound
	 * 
	 */
	//void giveTurretAngle(float angle);
	
	/* Name: setTurretBroken
	 * Arguments: true: broken
	 * 			  false: working
	 * Actions: Marks the turret encoder as broken
	 * Returns: nothin'
	 */
	void setTurretBroken(bool);
	
	/* Name: moveTurret
	 * Arguments: velocity: of turret, counterclockwise positive to forward main axis
	 * 						-1: clockwise full power, 1: ccws full power
	 * Actions: rotates the turret
	 * Returns: nothin'
	 */	
	void moveTurret(float rotvel); 

	/* Name: stopTurret
	 * Arguments: NONE!!!
	 * Actions: stops the turret from moving
	 * Returns: Nothing. Nothing at all.
	 */
	void stopTurret();
	
	/* Name: turretFinished
	 * Arguments: none
	 * Actions: checks if turret is close enough to target position
	 * Returns: true: if successful
	 */
	bool turretFinished();

	/* Name: getTurretAngle
	 * Arguments: none
	 * Actions: find angle of turret
	 * Returns: degrees of above, positive being counterclockwise
	 * 			to robot forward axis
	 */
	float getTurretAngle();// 
	
    /* Name: getTurretDesiredAngle
     * Arguments: none
     * Returns: desired angle of turret
     */
    float getTurretDesiredAngle();// 

    /* Raw turret encoder value */
	int getTurretEnc();
	
    /* normalized turret encoder value */
    double getTurretEncNorm();
    
    /* Gets error percentage of turret (-1 -> 1 where 1 is 100% error) */
    double getTurretError();
    
	//Sets a flag to cause the turret to zero itself
	//Will override any input from move commands while zeroing
	void requestZero();
	
	//Processes turret functions, including zeroing
	void processTurret();
	
	//Returns allowed speed based on limit switches
	float checkLimits(float requestSpeed);
	
	//returns the voltages from the limit sensors on the turret
	float getLimitCW();
	float getLimitCCW();
	
	// Turn on/off the logging of shooter PID. Does nothing if
	// shooter logging is disabled
	void setLoggingEnabled(bool en);
	
protected: // SmartShooter NEEDS TO KNOW	
	bool wheels_broken;
	bool broken_turret; // is turret broken
		
private: // SmartShooter doesn't need to know
	Storage * store;
	
	ThunderCANJaguar * turret_jaguar;
	ThunderCANJaguar * top_jaguar;
	ThunderCANJaguar * bottom_jaguar;
	
	Encoder * turret_encoder;
	MonoEncoder * top_encoder;
	MonoEncoder * bottom_encoder;

#if SHOOTER_LOGGING
	UDPLogger *top_logger;
    UDPLogger *bottom_logger;
    LoggingPIDOutput *top_logger_output;
    LoggingPIDOutput *bottom_logger_output;
#endif

	LimitedOutput * turretOutput;
	
	PIDController * turret_PID;
	PIDController * top_PID;
	PIDController * bottom_PID;
	AnalogChannel * limitCW;
	AnalogChannel * limitCCW;
	Timer * shot_timer;
	Timer * speedup_timer;

	bool turret_move_manual;
	bool want_zero;//Do we want to zero the turret?
	
	bool want_shooting; // has shooting been called
	bool shooting; // is it shooting
	
	int top_at_speed;
	int bottom_at_speed;
	double at_speed_time;
	bool useBackspin;
	
	// Q: Sensors to limit turret? A: Sensors and a hard stop.
	// (read, rstop->angle when motion is impossible)
	float turret_rotvel;// rotational velocity for turret
	bool turret_to_position; // is the turret where it should be
	float turret_current_angle;
	float turret_target_angle;
	
	float top_speed;// wheel speed goal: FROM 0 to 1
	float bottom_speed;// wheel speed goal FROM 0 to 1

	bool shooter_enabled;// whether the shooter is enabled (wheels need to spin)
						// or not
	bool speed_variable_mode; // have variable wheel speeds or not
	
	void shooterEnable();// make the wheels spin or not depending
	void shooterDisable();// on the need to shoot
	
	bool turretOnTarget();
	void findZero();//Finds the CW limit and resets the encoder count to zero 

	bool isReady();// Checks if speed is achieved,
					//turret stable, ball available
};


#endif /*SHOOTER*/
