#ifndef STORAGE
#define STORAGE

#include "Relay.h"
#include "WPILib.h"
#include "ThunderCANJaguar.h"
#include "IOmap.h"

class Storage {
public:
	Storage();
	~Storage();
	
	typedef enum { WINGS_UP, WINGS_DOWN, WINGS_STOP, WINGS_HALF } WingsPosition ;
	
// Process --------------------------------------------------------
	
	/* Name: processStorage
	 * Arguments: nothin'
	 * Actions: Moves balls to next available position; controls beater bar, 
	 * 		conveyor belt, and rotator. If not all balls are there, allows entrance
	 * 		of more balls and moves them into appropriate positions. If all 3 balls
	 * 		are stored, repels balls. Also moves wings.
	 * Returns: Number of balls in the storage system
	 * 			
	 */
	int processStorage();
	
	/* Name: ballShot
	 * Arguments: none
	 * Actions: When a ball is shot, informs that one less ball is in storage
	 * Returns: nothing 
	 */
	void ballShot();

// Conveyors ------------------------------------------------------
	
	/* Name: getBallCount
	 * Arguments: nothin'
	 * Actions: Counts the # of balls in storage system
	 * Returns: # of balls
	 */
	int getBallCount();

	/* Name: isReady
	 * Arguments: nothin'
	 * Actions: checks if a ball is readily available to shoot
	 * Returns: 
	 * 		false:	if ball is not ready to shoot
	 *		true:	if ball is ready to shoot
	 */
	bool isReady();

	/* Name: dump
	 * Arguments: true: dump
	 * 			  false: not
	 * Actions: sets mode of system to sending all balls out
	 * Returns: nothing
	 */
	void dump(bool dump);
	
	/* Name: deliverBall
	 * Arguments: nothin'
	 * Actions: Delivers a ball from the final position in storage to the shooter.
	 * Returns: nothing
	 */
	void deliverBall();
	
	bool deliverInProgress();
	
	/* Stops delivering a ball if one is in the process of being delivered.
	 */
	void cancelDeliverBall();
	
	/* Name: convBeltXMove
	 * Arguments: -1: backwards
	 * 			   0: do nothing
	 * 			   1: forwards
	 * Actions: Move the conveyor belt, with forward moving a ball toward the shooter
	 * Returns: nothing
	 */
	void convBelt1Move(int);
	void convBelt2Move(int);
	void convBelt3Move(int);
	
	/* Name: manualConvMove
	 * Arguments:  -1: outwards
	 * 				1: inwards
	 * 				0: automatic
	 *              -2: stage1 only go out, other stages stop
	 * 				anything else: stop
	 */
	void manualConvMove(int dir);
	
	
	/* Name: setConvBroken
	 * Arguments:	true: broken
	 * 				false: working
	 * Actions: marks the system of sensors for the conveyor belts as incomplete
	 * Returns: nothing
	 */
	void setConvBroken(bool broken);
	
	/* Name: isConvBroken
	 * Arguments: none
	 * Actions: Finds if the conveyor belts are broken
	 * Returns: True: broken
	 * 			False: working
	 */
	bool isConvBroken();
	
	/* Name: setCountBroken
	 * Arguments: 	true: is broken
	 * 				false: working
	 * Actions: marks the first sensor, the one used to count balls at the storage entrance, as broken
	 * Returns: nothing
	 */
	void setCountBroken(bool broken);
	
	/* Name: isCountBroken
	 * Arguments: none
	 * Actions: Finds if the sensor at storage entrance is broken
	 * Returns: True: Broken
	 * 			False: Working
	 */
	bool isCountBroken();
	

	float getSensorVoltage(int sensor);
    bool getSensorState(int sensor);

	
	
// Wings -----------------------------------------------------------
	
	/* Name: set wing position using sensors to achieve the given position
	 *       Has no effect if wings are set broken
	 *       For down, tries to achieve the given direction for a limited time,
	 *       then gives up until commanded to go the other direction.
	 *       For up, goes up and tries to stay up by driving as necessary.
	 * Arguments: true: down
	 * 			  false: up
	 * Actions: Sets position of the wings
	 * Returns: nothin'
	 */
	void setWingsPosition(WingsPosition pos);// inc. timeout thing, move action to process
	
    /* Name: wingsMove
     * Arguments: -1: upwards
     *             0: stop
     *             1: downwards
     * Actions: Move the wings- moves wings manually in the given direction and
     *          an internal, preset speed. Overrides and terminates any previous
     *          instruction given to setWingsPosition().
     * Returns: nothing
     */
    void wingsMove(int dir);
    
	/* Name: atWingsPosition
	 * Arguments: none
	 * Actions: checks if wings are up or down
	 * Returns: true: if the wings are up when they were told to be up
	 * 			true: if wings are down when told so
	 *          true: if timeout occurred while moving the wings
	 * 			false: otherwise
	 */
	bool atWingsPosition();
	
	/* Name: wingsStop
	 * Arguments: none
	 * Actions: Stops any motion of the wings
	 * Returns: nothing
	 */
	void wingsStop();
	
	/* Name: setWingsBroken
	 * Arguments:	true: broken
	 * 				false: working
	 * Actions: marks the wing sensors as broken or not
	 * Returns: nothing
	 */
	void setWingsBroken(bool broken);
	
	/* Name: areWingsBroken
	 * Arguments: none
	 * Actions: checks to see if the wings sensors are broken
	 * Returns: True: broken
	 * 			False: working
	 */
	bool areWingsBroken();

	/*Name: getWingsPosition
	 * Arguments: none
	 * Actions: Gets the current position of the wings
	 * Returns: -1: bottom
	 * 			1: top
	 * 			0: none of the above (broken/stopped in between)
	 * 			2: wings halfway down
	 */
	int getWingsPosition();
	
    /* Name: wingsPos
     * Arguments; none
     * Action: Returns the output of the wings
     * Returns: A normalized number from 0-1 as the position of the wings
     *          where 0 means full deployed and 1 means full retract
     */
    float wingsPos();
    
	/* Name: wingsRawValue
	 * Arguments; nuttin
	 * Action: Gets the raw voltage of the wings position
	 * Returns the voltage of the wings positon
	 */
	float wingsRawValue();

// Beater Bar -------------------------------------------------------

	/* Name: beaterBarMove
	 * Arguments: -1: backwards (balls sent out)
	 * 			   0: run automatically
	 * 			   1: forwards (balls in)
	 *             Anything else: Turn it off and don't move
	 * Actions: manually turns beater bar
	 * Returns: nothing
	 */
	void beaterBarMove(int dir);
	
	/* Name: getBeaterBarMode
	 * Arguments: nothin'
	 * Actions: returns the direction the beater bar is moving
	 * Returns:   1: Forwards / taking in balls
     *            0: Auto movement
     *           -1: Backwards / repelling balls
     *           Anything else: Manually stopped
	 */
	int getBeaterBarMode();
	
	// Bridge Stabilizer -------------------------------------------------------

	// Activate the bridge stabilizer.
	// + values spin in deploy direction, 0 is stop, - values unwind
	void setBridgeStabilizer(float speed);
	
	// Get state of bridge stabilizer's sensor.  True of the stabilizer is
	// locked correctly in the up position (as in, not fired)
	bool isBridgeStabilizerLocked();

private:
	Relay * convBelt1;
	Relay * convBelt2;
	Relay * convBelt3;
#if ROBOT_2
	Relay * beaterBar;
#else
	Victor * beaterBar;
#endif
	Victor * bridgeStabilizer; // Not a Victor - Vex motor
	DigitalInput * bridgeStabilizerUnlocked;
	ThunderCANJaguar * wings;
	AnalogChannel * wingsPosition;
	DigitalInput * countSensor;
	AnalogChannel * convPos1;
	AnalogChannel * convPos2;
	AnalogChannel * convPos3;
	
	Timer * wingsTimer;
	
	//Sensors to limit wings
	
	WingsPosition wingsDesiredPos;
	float wingsManualSpeed;
	float wingsAutoDownSpeed;
	float wingsLastPos;
	bool wingsAtTop;
	
	enum { BEATER_IN, BEATER_OUT, BEATER_STOP, BEATER_AUTO } beaterBarMode;
	Timer *beaterBarChangeTimer;            // Time beater bar has been off
	                                        // invalid value if beaterBarVal != Relay::kOff
    Relay::Value beaterBarLastDir;          // Last non-off direction
	Relay::Value beaterBarValue;            // Last output to beaterBar relay
    void beaterBarSet(Relay::Value val);
	
	enum { 	CONV_IN, CONV_OUT, CONV_AUTO, CONV_STOP, CONV_OUT_1 } conveyorBeltMode;
	
	bool deliver;// if ball is being sent to be shot
	bool conv_sensors_broken;// the sensors to stop conveyors. if one is gone, go manual
	bool count_sensors_broken;// the initial sensor
	

	bool wingSensorsBroken;// If the sensors for the wings are broken
	
	bool shootLimit; // Ends movement of conveyor belt once ball is shot
	bool sensor3_last; // Ends movement of conveyor belt once ball is shot
	
	// sensorY: senses [if a ball is in the way of a line]
    bool sensor0();//First sensor, trips when ball enters just enough
                // not to require beater bar assistance
    bool sensor1();// Second sensor,trips when ball is at desired position

    bool sensor2();// Trips when ball is at desired position
    
    bool sensor3();// Trips when ball is at desired position
    
};

#endif /*STORAGE*/
