#include <math.h>
#include "ThunderDrive.h"
#include "IOmap.h"

// Controller factors for driving (movement)
// Want output to be 1.0 (max power) when distance >= 20 inches, so P = 1/20  
#define DRIVE_P  (1.0/20.0)
// Minimum power to motors during auto drive
#define DRIVE_MIN 0.25
// auto drive acceleration rate - motor power per second
#define DRIVE_ACCEL 1.0
// Multiplicative factor applied to angular error when driving straight
// Amount of motor power boost per degree of error
#define KEEP_STRAIGHT_CORRECTION_FACTOR 0.05

// Controller factors for turning
// Want output to be 1.0 (max turn) when error >= 45 degrees, so P = 1/45 
#define TURN_P  (1.0/45.0)
#define TURN_MIN 0.25

ThunderDrive* ThunderDrive::theDrivep = 0;

// Driver station sends data fifty times/second (every 20 ms)
// If we miss several, we've lost contact.  Hit the brakes!
#define SAFETY_TIMEOUT_SECS (3 * 0.020)

// Singleton function returns ptr to one and only ThunderDrive object
// Only creates one the first time it is called.  After that, it returns
// a pointer to that same one.
ThunderDrive* ThunderDrive::getInstance()
{
	if(theDrivep == 0)
		theDrivep = new ThunderDrive;
	return theDrivep;
}

// Constructor (private) - only used by getThunderDrivePtr()
ThunderDrive::ThunderDrive() :
			lfmotor(CANID_DRIVETRAIN_LEFT_FRONT),
			lrmotor(CANID_DRIVETRAIN_LEFT_REAR),
			rfmotor(CANID_DRIVETRAIN_RIGHT_FRONT),
			rrmotor(CANID_DRIVETRAIN_RIGHT_REAR),
			lencoder(DIG_DRIVETRAIN_LEFT_ENCODER_A),
			rencoder(DIG_DRIVETRAIN_RIGHT_ENCODER_A),
			gyro(ANA_DRIVETRAIN_GYRO),
			autoState(AUTO_NONE)
{
	bool motorSafety = false;
	lfmotor.SetSafetyEnabled(motorSafety);
	lrmotor.SetSafetyEnabled(motorSafety);
	rfmotor.SetSafetyEnabled(motorSafety);
	rrmotor.SetSafetyEnabled(motorSafety);

	// Set default motor neutral mode
	neutralMode = ThunderCANJaguar::kNeutralMode_Brake;
	rfmotor.ConfigNeutralMode(neutralMode);
	rrmotor.ConfigNeutralMode(neutralMode);
	lfmotor.ConfigNeutralMode(neutralMode);
	lrmotor.ConfigNeutralMode(neutralMode);

	// Zero the sensors
	lencoder.Reset();
	lencoder.SetReverseDirection(true);
	lencoder.Start();
	rencoder.Reset();
	rencoder.SetReverseDirection(false);
	rencoder.Start();

	// Gyro is a http://www.analog.com/static/imported-files/data_sheets/ADXRS652.pdf
    // 7 mv/Deg/Sec
	gyro.SetSensitivity(0.007f);
	gyro.Reset();

#if SOFT_START
	softStartTimer.Start();
#endif
	
	autoTimer = new Timer();
	
	lastLeftPower = 0.0;
	lastRightPower = 0.0;

	debugOutTimer.Start();

}

// destructor - should never e called
ThunderDrive::~ThunderDrive()
{
}

void ThunderDrive::processDrive()
{
    switch (autoState) {
    case AUTO_DRIVE:
    {
        int distLeft = autoTargetDistance - getEncoderAvg();
    	//printf("processDrive - distLeft:%i, getEncoderAvg:%i\n", distLeft, getEncoderAvg());
        if (distLeft <= DISTANCE_CLOSE_ENOUGH) {
            setMotorSpeedInternal(0, 0);
            autoState = AUTO_NONE;
        	printf("processDrive done - distLeft:%i, getEncoderAvg:%i\n", distLeft, getEncoderAvg());
        } else {
            double pPow = distLeft * DRIVE_P;
            double accelPow = autoTimer->Get() * DRIVE_ACCEL;
            if (pPow > autoMaxPower)
                pPow = autoMaxPower;
            
            // pPow is forward-motion power. Now gyro correct
            // negative delta means robot pointed left-of target
            double delta = getGyro() - autoTargetAngle;
            delta *= KEEP_STRAIGHT_CORRECTION_FACTOR;
            // Swap delta direction -- negative in driveAndTurn means turn left
            delta *= -1.0f;

            driveWithCorrection(pPow, -delta, DRIVE_MIN, accelPow);
        }
        break;
    }
    case AUTO_TURN:
    {
        // negative delta means robot pointed left-of target
        double delta = getGyro() - autoTargetAngle;
        if (fabs(delta) < ANGLE_CLOSE_ENOUGH) {
            setMotorSpeedInternal(0, 0);
            autoState = AUTO_NONE;
        } else {
            double turnPow = delta * TURN_P;
            if (turnPow < 0 && turnPow > -TURN_MIN)
                turnPow = -TURN_MIN;
            if (turnPow > 0 && turnPow < TURN_MIN)
                turnPow = TURN_MIN;
            // Negative turnpow means turn to right ->
            //     so push left positive and right negative
            // Positive turnpow means turn left
            //     so push left negative and right positive
            // Therefore, invert left
            setMotorSpeedInternal(-turnPow, turnPow);
        }
        break;
    }
    case AUTO_DRIVE_TURN:
        // Not yet implemented
        autoState = AUTO_NONE;
        break;
    default:
        // Do nothing - manual is handled directly
        break;
    }
}

//setMotorSpeed - Set motor speeds for teleoperated mode
void ThunderDrive::setMotorSpeed(float leftPower, float rightPower)
{
    setMotorSpeedInternal(leftPower, rightPower);
    autoState = AUTO_NONE;
}



void ThunderDrive::setMotorSpeedInternal(float leftPower, float rightPower)
{
	// Make sure power is in the range 1.0 to -1.0
	if (rightPower > 1.0)
		rightPower = 1.0;
	else if (rightPower < -1.0)
		rightPower = -1.0;
	if (leftPower > 1.0)
		leftPower = 1.0;
	else if (leftPower < -1.0)
		leftPower = -1.0;

#if SOFT_START
	// limit acceleration to reduce jerkiness and loss of orientation
	// caused by zoom starting

	//Are we accelerating?
	if((fabs(rightPower)> fabs(lastRightPower))
			|| (fabs(leftPower)> fabs(lastLeftPower)))
	{
		float elapsedTime = softStartTimer.Get();
		float maxChangeAllowed = ACCELERATION_LIMIT * elapsedTime;

		float rightChange = rightPower - lastRightPower; // + if fwd, - if reverse
		float leftChange = leftPower - lastLeftPower; // + if fwd, - if reverse
		if(fabs(rightChange)> maxChangeAllowed)
		{
			printf("ss right\n");
			rightPower = lastRightPower + maxChangeAllowed * fSignOf(rightPower);
		}
		if(fabs(leftChange)> maxChangeAllowed)
		{
			printf("ss left\n");
			leftPower = lastLeftPower + maxChangeAllowed * fSignOf(leftPower);
		}
	}
	softStartTimer.Reset();
#endif

	rfmotor.Set(-rightPower);
	rrmotor.Set(-rightPower);
	lfmotor.Set(leftPower);
	lrmotor.Set(leftPower);
	lastLeftPower = leftPower;
	lastRightPower = rightPower;

	return;
} // end ThunderDrive::setMotorSpeed

static int inchesToTicks(float inches) {
    return (int)(inches * ENCODER_COUNTS_PER_INCH);
}

//autonomousMove - begins a straight line movement a certain distance in inches
void ThunderDrive::autonomousMove(float distance, float maxPower)
{

	printf("autonomousMove - dist:%f, maxSpeed:%f\n", distance, maxPower);
	
	 // Cancel any move or turn actions in progress
	stopAction();

	autoState = AUTO_DRIVE;
	autoMaxPower = maxPower;
	if (distance > 0) {
	    autoDirection = -1.0f;
	} else {
	    distance = -distance;
	    autoDirection = 1.0f;
	}
    autoTargetDistance = getEncoderAvg() + inchesToTicks(distance);
	autoTargetAngle = getGyro();
	printf("autonomousMove - getEncoderAvg:%i targetdist:%i, targetangle:%f\n", getEncoderAvg(), autoTargetDistance, autoTargetAngle);
	autoTimer->Reset();
	autoTimer->Start();

	return;
} // end ThunderDrive::autonomousMove

//autonomousTurn - begins a turn in place a certain angle from the starting direction
void ThunderDrive::autonomousTurn(float degrees)
{
	stopAction(); // Cancel any actions in progress	
	
    autoState = AUTO_TURN;
    autoMaxPower = 1.0;
    autoTargetDistance = getEncoderAvg() + 0;
    autoTargetAngle = getGyro() + degrees;
    autoTimer->Reset();
    autoTimer->Start();
	
	return;
}

//autonomousDrivingTurn - begins a turn in place a certain angle from the starting direction
void ThunderDrive::autonomousDrivingTurn(float degrees, float distance, float power)
{
	stopAction(); // Cancel any actions in progress
	
    autoState = AUTO_DRIVE_TURN;
    autoMaxPower = power;
    if (distance < 0) {
        distance = -distance;
        autoDirection = -1.0f;
    } else {
        autoDirection = 1.0f;
    }
    autoTargetDistance = getEncoderAvg() + inchesToTicks(distance);
    autoTargetAngle = getGyro() + degrees;    
    autoTimer->Reset();
    autoTimer->Start();
	
	return;
}

//autonomousSimpleDriveTurn - Turns a specified number of degrees while driving wheels on
//                            only one side.  Turn radius is whatever the physics dictates.
void ThunderDrive::autonomousSimpleDriveTurn(float degrees, float power)
{
	stopAction(); // Cancel any actions in progress	
    autoState = AUTO_TURN;
    autoMaxPower = power;
    autoTargetDistance = getEncoderAvg() + 0;
    autoTargetAngle = getGyro() + degrees;
    autoTimer->Reset();
    autoTimer->Start();

    return;
}

//doneYet - tells whether last called autonomous command was completed
bool ThunderDrive::doneYet()
{
	return autoState == AUTO_NONE;
}

//void stopAction - stops all drive and turn actions
void ThunderDrive::stopAction()
{
	autoState = AUTO_NONE;
	setMotorSpeedInternal(0, 0);

	return;
}

//int getEncoder - returns the left or right encoder reading
INT32 ThunderDrive::getEncoder(bool rightMotor)
{
	if (rightMotor)
		return rencoder.Get();
	else
		return lencoder.Get();
}

//int getEncoderAvg - returns the average of the left and right encoder readings, in counts
INT32 ThunderDrive::getEncoderAvg()
{
	return (rencoder.Get() + lencoder.Get()) / 2;
}

// getGyro - retrieves the current gyro value, in degrees.
float ThunderDrive::getGyro(void)
{
	return gyro.GetAngle();
}

// setCoast - Coast or brake when motor drive is zero?
void ThunderDrive::setCoast(bool coast)
{
	if (coast)
		neutralMode = ThunderCANJaguar::kNeutralMode_Coast;
	else
		neutralMode = ThunderCANJaguar::kNeutralMode_Brake;

	rfmotor.ConfigNeutralMode(neutralMode);
	rrmotor.ConfigNeutralMode(neutralMode);
	lfmotor.ConfigNeutralMode(neutralMode);
	lrmotor.ConfigNeutralMode(neutralMode);
	return;
}

// driveAndTurn - Set motor speeds to drive and turn in autonomous mode
void ThunderDrive::driveWithCorrection(float power, float turn, float min, float max)
{
	// IF turn == 0, left and right power should both equal the power value
	// IF power == 0 and turn == 0, left and right power should be zero
	// IF power == 0 and turn = -1, (full left turn in place) right power should be -1 and left power should be +1
	// IF power == 0 and turn = 1, (full right turn in place) left power should be -1 and right power should be +1
	// BUT IF power = 1 and turn = 1 (full speed right turn) left=0, right =1
	
	// Start with each equal to power
	float left = power;
	float right = power;
	
	// Add/subtract turn value
	left = left - turn;
	right = right + turn;
	
	// Clamp
    if (left > max) left = max;
    else if (left < min) left = min;    
    if (right > max) right = max;
    else if (right < min) right = min;
	
	// Apply the computed power to the drive motors
	setMotorSpeedInternal(autoDirection * left, autoDirection * right); 
}

// Return sign of f as either +1.0 or -1.0
float ThunderDrive::fSignOf(float f)
{
	if (f >= 0.0)
		return 1.0;
	else
		return -1.0;
}

