#ifndef THUNDERDRIVE_H
#define THUNDERDRIVE_H


#include "WPILib.h"
#include "ThunderCANJaguar.h"

// Set SOFT_START to 1 to enable acceleration limitation
#define SOFT_START (0)

#if SOFT_START
// Acceleration limit in drive per second.
// For example, a value of 1 means that increasing the drive
// instantly from 0.0 to 1.0 will result in the robot taking
// 1 second to take the motors from 0.0 to 1.0 drive.
// 0.1 mans ten seconds from 0 to full
#define ACCELERATION_LIMIT (1)
#endif

#define ENCODER_COUNTS_PER_INCH (29)

// How close to setpoint to be considered done?
// In inches
#define DISTANCE_CLOSE_ENOUGH (0.5)
// In degrees
#define ANGLE_CLOSE_ENOUGH (1.0)

class ThunderDrive
{
public:


	// getInstance - Static function returns a pointer to the one and only ThunderDrive 
	// object in the system.
	// Use it like this:
	//
	//      ThunderDrive* tdp; pointer to the system ThunderDrive object;
	//      tdp = ThunderDrive::getInstance();
	//      .  .  .
	//      tdp->setMotorSpeed(. . .);
	
	static ThunderDrive* getInstance();
	
    ~ThunderDrive();

    
    void processDrive();

	//setMotorSpeed - Set motor speeds for teleoperated mode
	//        Inputs:
	//                leftPower: Desired power for left motor, range: -1.0 -> 1.0
	//                rightPower: Desired power for right motor, range: -1.0 -> 1.0
	//        Returns: nothin'
	void setMotorSpeed(float leftPower, float rightPower);
	
	//autonomousMove - begins a straight line movement a certain distance in inches
	//        Inputs:
	//                distance: the distance in inches that the robot must move
	//                        positive values: forwards
	//                        negative values: backwards
	//                maxPower: max power value for the motors
	//        Returns: Nothin'
	//        Note: If a previous autonomousMove() operation is under way, resets stopping point to "distance" from the present point
	//            If a previous autonomousTurn() operation is under way, cancels the turn, and then moves
	//            doneYet() function used to check completion
	void autonomousMove(float distance, float maxPower);

	//autonomousTurn - begins a turn in place a certain angle from the starting direction
	//        Inputs:
	//                degrees: The angle at which to turn, in degrees
	//                        positive value: turn to the right
	//                        negative value: turn to the left
	//        Returns: Nothing
	//        Note: The same applies to autonomousTurn as does autonomousMove. If a previous operation of autonomousTurn
	//            or autonomousMove is still in progress with autonomousTurn is called, it will interrupt the current
	//            action and start a new autonomousTurn from the location and angle it is currently at.
	void autonomousTurn(float degrees);

	//autonomousDrivingTurn - begins a turn in place a certain angle from the starting direction
	//        Inputs:
	//                degrees: The angle at which to turn, in degrees
	//                        positive value: turn to the right
	//                        negative value: turn to the left
	//				   distance: distance in inches that the robot moves while turning
	//                 maxPower: The maximum speed for the move, 0 -> 1.0
	//        Returns: Nothing
	//        Note: The same applies to autonomousDrivingTurn as does autonomousMove. If a previous operation of autonomousTurn
	//            or autonomousMove is still in progress with autonomousTurn is called, it will interrupt the current
	//            action and start a new autonomousTurn from the location and angle it is currently at.
	void autonomousDrivingTurn(float degrees, float distance, float maxPower);

	//autonomousSimpleDriveTurn - Turns a specified number of degrees while driving wheels on
	//                            only one side.  Other side freewheels.  Turn radius is whatever
	//                            the physics dictates.
	//        Inputs:
	//                degrees: The angle at which to turn, in degrees
	//                        positive value: turn to the right
	//                        negative value: turn to the left
	//                 maxPower: maximum power to apply to the motor that is running
	//        Returns: Nothing
	//        Note: The same applies to autonomousSimpleDriveTurn as does autonomousMove. If a previous operation of autonomousTurn
	//            or autonomousMove is still in progress with autonomousTurn is called, it will interrupt the current
	//            action and start a new autonomousTurn from the location and angle it is currently at.
	void autonomousSimpleDriveTurn(float degrees, float maxPower);

	//doneYet - tells whether last called autonomous command was completed
	//            Inputs: none
	//            Returns:
	//                    True: Last command was completed
	//                    False: Last command is in process of completion
	//            Note: This can be used to check the result of the autonomousMove() and autonomousTurn() commands
	bool doneYet();

	/* getEncoder - retrieves the current encoder value.
	 * 			Inputs: bool rightMotor
	 * 				true: returns the encoder value for the right motor
	 * 				false: returns the encoder value for the left motor
	 * 					If no value is given, the function defaults to the right motor
	 * 			Returns: The number of ticks, as an INT32. 	
	 * 			Note: The encoder is reset ONCE when the ThunderDrive object is instantiated.
	 */
	INT32 getEncoder(bool rightMotor = true);

	/* getEncoderAvg - retrieves the average of the left and right encoder value.
	 * 			Returns: The average number of ticks, as an INT32. 	
	 * 			Note: The encoder is reset ONCE when the ThunderDrive object is instantiated.
	 */
	INT32 getEncoderAvg();

	/* getGyro - retrieves the current gyro value.
	 * 			Inputs: bool rightMotor
	 * 				true: returns the encoder value for the right motor
	 * 				false: returns the encoder value for the left motor
	 * 					If no value is given, the function defaults to the right motor
	 * 			Returns: Heading in degrees, as a float. 					
	 *          Note: The gyro is reset ONCE when the Thunderdrive object is instantiated. 
	 */
	float getGyro(void);

	// setCoast - Coast or brake when motor drive is zero?
	//          Inputs: coast
	//                        ciast if true, brake if false 
	void setCoast(bool coast);

	/*void stopAction - signals to end current action
	 * 		Inputs:
	 * 				Nothin'
	 * 		Returns: Nothin' 								*/
	void stopAction();


private:
	//ThunderDrive's constructor - private to force people to use getThunderDrivePtr()
	ThunderDrive();

	// driveWithCorrection - Set motor speeds to drive and turn in autonomous mode
	//        Inputs:
	//                power: Desired power for motors, range: 0.0 -> 1.0
	//                turn:  Desired turn, range: -1.0 -> 1.0
	//                          -1.0 means max counterclockwise turn
	//                           0.0 means no turn
	//                          +1.0 means max clockwise turn
	//                min:   Minimum value for each motor
	//                max:   Maximum value for each motor
	//        Returns: nothin'
	//        Notes: min and max are specified in absolute value terms.
	//               max always wins when min/max overlap
	//               combined power and turn get autoDirection applied to them
	//                  after clamping but before output
	void driveWithCorrection(float power, float turn, float min, float max);
	
	// Clamps, inverts if needed, and outputs the specified motor speeds
	void setMotorSpeedInternal(float left, float right);

	float fSignOf(float f);

	// Singleton pointer to the one and only ThunderDrive object in the system
	static ThunderDrive* theDrivep;
	
	// Motors
	ThunderCANJaguar lfmotor;
	ThunderCANJaguar lrmotor;
	ThunderCANJaguar rfmotor;
	ThunderCANJaguar rrmotor;

	//Sensors
	Counter lencoder; // left wheel encoder
	Counter rencoder; // right wheel encoder
	Gyro gyro;

	Timer debugOutTimer; // When do we print debug msg?


	///////////////////////////////////////////////////////////////////////////////////////
	
	ThunderCANJaguar::NeutralMode neutralMode; // What do motors do when power = 0?
	
	// Autonomous drive state & support
	enum { AUTO_NONE, AUTO_DRIVE, AUTO_TURN, AUTO_DRIVE_TURN } autoState;
	float autoMaxPower;
	int autoTargetDistance;          // Ticks
	float autoTargetAngle;             // Degrees
	float autoDirection;               // -1 rev or 1 fwd
	Timer *autoTimer;
	
	// Last power values we sent to the motors
	float lastLeftPower;
	float lastRightPower;

#if SOFT_START
	Timer softStartTimer;
#endif

}; // end ThunderDrive 

#endif /*THUNDERDRIVE_H*/
