////////////////////////////////////////////////////////////////////////////////////////////////////
//
//	The order for processing an image:
//		1. ConvertRgbToMono			pulls the green from the camera image and creates monochrome
//										image containing just the green targets
//		2. CalcTargetRows			uses mono image to determine vertically where the targets are
//		3. CalcTargetColumns		uses mono image and CalcTargetRows() results to determine
//										horizontally where each target is
//		4. CalcTargets				uses mono image and CalcTargetColumns() results to 'tighten'
//										the boxes for each target and calc final values
//		OR
//		ProcessImage				will process the entire image all at once (i.e. run STEPs 1 - 4)
//
//	After processing:
//		GetTargetInfo				gets list of targets found
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef THUNDERVISION
#define THUNDERVISION

#include "WPILib.h"

class ThunderVision
{
public:
	ThunderVision();
	~ThunderVision();

	// information we need to track for each target
	struct TARGET_INFO
	{
		int row;						// row within the image (0 is first row)
		int left;						// left-most pixel
		int right;						// right-most pixel
		int top;						// top-most pixel
		int bottom;						// bottom-most pixel
		int avgWidth;					// average width
		int avgHeight;					// average height
		int maxWidth;					// max width found
		int maxHeight;					// max height found
		int edgeLeft;					// 'angle' of left edge (0 = straight; neg = \; pos = /)
		int edgeTop;					// 'angle' of top edge (0 = straight; neg = /; pos = \)
	};

	// will process the entire image all at once (i.e. run STEPs 1 through 4)
	//
	// returns the number of targets found
	int ProcessImage(RGBValue *image);

	// convert the given RGB image (from the camera) into a monochrome image, by pulling
	// out the green tape reflections. the mono image is stored within this class
	//
	// the image MUST be 640 x 480
	void ConvertRgbToMono(RGBValue *image);			// STEP 1

	// look at each row to see if it has any pixels in it.
	// any successive lines (with pixels) will be considered part of the same target
	void CalcTargetRows();							// STEP 2

	// look at each row (from CalcTargetRows()) and find all each box
	void CalcTargetColumns();						// STEP 3

	// look at each target from CalcTargetColumns() and 'fine tune' it and get final values
	void CalcTargets();								// STEP 4

	// gets the targets
	//
	// targetInfo		is a pointer to a list of targets (pass in NULL to get the total)
	// numEntires		indicates how many TARGET_INFO structures that targetInfo points to
	// returns			the number of targets copied
	int GetTargetInfo(TARGET_INFO *targetInfo, int numEntries);

	// for PC test app only
	#ifdef PC_TEST_APP
	void GetMonoImage(unsigned char *image) { memcpy(image, monoImage, sizeof(monoImage)); }
	void GetTargetRowImage(RGBValue *image);		// uses CalcTargetRows() results
	void GetTargetImageWithGap(RGBValue *image);	// uses CalcTargetColumns() results (with gaps)
	void GetTargetImageNoGap(RGBValue *image);		// uses CalcTargetColumns() results (w/o gaps)
	void GetTargetImageNoNoise(RGBValue *image);	// uses CalcTargetColumns() results (final w/o gaps, w/o noise)
	void GetTargetImageTight(RGBValue *image);		// uses CalcTargets() result (tight box)
	#endif // PC_TEST_APP

private:

	// defines for pulling the green from the RGB camera image
	#define GREEN_THRESHOLD	110		// minimum value Green must be
	#define RED_DELTA		50		// Red must be this far away from Green [ignore if (r >= (g-RED_DELTA))]
	#define BLUE_THRESHOLD	40		// minimum value Blue must be
	#define BLUE_DELTA		10		// Blue cannot be this much brigher than Green [ignore if (b > (g+BLUE_DELTA))]

	#define MIN_PIXELS_ROW	5		// minimum pixels-per-row we need (use to ignore diagonal lines and noise)

	struct ROW_INFO							// information we need to track for each row
	{
		int top;
		int bottom;
	};

	// initialize member variables
	void Init();

	// create column histogram for the given target row
	void CreateRowHisto(int row);

	// look at row's histogram and find all the targets (just getting 'width' of target)
	void FindTargets(int targetFirstInRow, int gap);

	// use CalcTargetColumns() info to fine tune the top & bottom, and get avgWidth
	void FineTune(TARGET_INFO *targetInfo);
	void CalcHeightAndTopEdge(TARGET_INFO *targetInfo);


	unsigned char monoImage[640 * 480];		// grayscale image
	int histogram[640];						// histogram for width and height

	#define MAX_ROWS	8					// max rows we will track (should only be 3, but track more for noise/debug)
	ROW_INFO rowInfo[MAX_ROWS];				// all row info
	int rows;								// number of rows that have targets (i.e. valid items in rowInfo[])

	#define MAX_TARGETS	16					// max targets we will track (should only be 4, but track more for noise/debug)
	TARGET_INFO targetInfo[MAX_TARGETS];	// all target info
	int targets;							// number of targets (i.e. valid items in targetInfo[])

	// for PC test app only
	#ifdef PC_TEST_APP

	void GetTargetImage(RGBValue *image, int maxTargets, TARGET_INFO *targetInfo);

	#define MAX_TARGETS_TEST	64			// list of targets before we 'remove' the gaps
	TARGET_INFO targetInfoTest[MAX_TARGETS_TEST];
	int targetsTest;

	#define MAX_TARGETS_TEST_G	64			// list of targets w/o gaps but w/noise
	TARGET_INFO targetInfoTestG[MAX_TARGETS_TEST_G];
	int targetsTestG;

	#endif // PC_TEST_APP
};

#endif	// THUNDERVISION
